/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `BigQueryDriver` and related types declaration.
 */
import { Required } from '@cubejs-backend/shared';
import { BigQuery, BigQueryOptions, Dataset, Job, QueryRowsResponse } from '@google-cloud/bigquery';
import { Bucket, Storage } from '@google-cloud/storage';
import { BaseDriver, DriverCapabilities, DriverInterface, QueryColumnsResult, QueryOptions, QuerySchemasResult, QueryTablesResult, StreamTableData, TableCSVData } from '@cubejs-backend/base-driver';
import type { Query } from '@google-cloud/bigquery/build/src/bigquery';
interface BigQueryDriverOptions extends BigQueryOptions {
    readOnly?: boolean;
    projectId?: string;
    keyFilename?: string;
    exportBucket?: string;
    location?: string;
    pollTimeout?: number;
    pollMaxInterval?: number;
    /**
     * The export bucket CSV file escape symbol.
     */
    exportBucketCsvEscapeSymbol?: string;
}
type BigQueryDriverOptionsInitialized = Required<BigQueryDriverOptions, 'pollTimeout' | 'pollMaxInterval'>;
/**
 * BigQuery driver.
 */
export declare class BigQueryDriver extends BaseDriver implements DriverInterface {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency(): number;
    protected readonly options: BigQueryDriverOptionsInitialized;
    protected readonly bigquery: BigQuery;
    protected readonly storage: Storage | null;
    protected readonly bucket: Bucket | null;
    /**
     * Class constructor.
     */
    constructor(config?: BigQueryDriverOptions & {
        /**
         * Data source name.
         */
        dataSource?: string;
        /**
         * Max pool size value for the [cube]<-->[db] pool.
         */
        maxPoolSize?: number;
        /**
         * Time to wait for a response from a connection after validation
         * request before determining it as not valid. Default - 10000 ms.
         */
        testConnectionTimeout?: number;
    });
    static driverEnvVariables(): string[];
    testConnection(): Promise<void>;
    readOnly(): boolean;
    query<R = unknown>(query: string, values: unknown[], options?: QueryOptions): Promise<R[]>;
    protected loadTablesForDataset(dataset: Dataset): Promise<any>;
    tablesSchema(): Promise<any>;
    getSchemas(): Promise<QuerySchemasResult[]>;
    getTablesForSpecificSchemas(schemas: QuerySchemasResult[]): Promise<QueryTablesResult[]>;
    getColumnsForSpecificTables(tables: QueryTablesResult[]): Promise<QueryColumnsResult[]>;
    getTablesQuery(schemaName: string): Promise<{
        table_name: string | undefined;
    }[]>;
    tableColumnTypes(table: string): Promise<any>;
    createSchemaIfNotExists(schemaName: string): Promise<void>;
    isUnloadSupported(): Promise<boolean>;
    stream(query: string, values: unknown[]): Promise<StreamTableData>;
    unload(table: string): Promise<TableCSVData>;
    loadPreAggregationIntoTable(preAggregationTableName: string, loadSql: string, params: any, options: any): Promise<any>;
    protected awaitForJobStatus(job: Job, options: any, withResults: boolean): Promise<true | QueryRowsResponse | null>;
    protected runQueryJob<T = QueryRowsResponse>(bigQueryQuery: Query, options: any, withResults?: boolean): Promise<T>;
    protected waitForJobResult(job: Job, options: any, withResults: boolean): Promise<true | QueryRowsResponse>;
    quoteIdentifier(identifier: string): string;
    capabilities(): DriverCapabilities;
}
export {};
//# sourceMappingURL=BigQueryDriver.d.ts.map