"use strict";
/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `BigQueryDriver` and related types declaration.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BigQueryDriver = void 0;
const shared_1 = require("@cubejs-backend/shared");
const ramda_1 = __importDefault(require("ramda"));
const bigquery_1 = require("@google-cloud/bigquery");
const storage_1 = require("@google-cloud/storage");
const base_driver_1 = require("@cubejs-backend/base-driver");
const HydrationStream_1 = require("./HydrationStream");
/**
 * BigQuery driver.
 */
class BigQueryDriver extends base_driver_1.BaseDriver {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 10;
    }
    /**
     * Class constructor.
     */
    constructor(config = {}) {
        super({
            testConnectionTimeout: config.testConnectionTimeout,
        });
        this.storage = null;
        this.bucket = null;
        const dataSource = config.dataSource ||
            (0, shared_1.assertDataSource)('default');
        this.options = {
            scopes: [
                'https://www.googleapis.com/auth/bigquery',
                'https://www.googleapis.com/auth/drive',
            ],
            projectId: (0, shared_1.getEnv)('bigqueryProjectId', { dataSource }),
            keyFilename: (0, shared_1.getEnv)('bigqueryKeyFile', { dataSource }),
            credentials: (0, shared_1.getEnv)('bigqueryCredentials', { dataSource })
                ? JSON.parse(Buffer.from((0, shared_1.getEnv)('bigqueryCredentials', { dataSource }), 'base64').toString('utf8'))
                : undefined,
            exportBucket: (0, shared_1.getEnv)('dbExportBucket', { dataSource }) ||
                (0, shared_1.getEnv)('bigqueryExportBucket', { dataSource }),
            location: (0, shared_1.getEnv)('bigqueryLocation', { dataSource }),
            ...config,
            pollTimeout: (config.pollTimeout ||
                (0, shared_1.getEnv)('dbPollTimeout', { dataSource }) ||
                (0, shared_1.getEnv)('dbQueryTimeout', { dataSource })) * 1000,
            pollMaxInterval: (config.pollMaxInterval ||
                (0, shared_1.getEnv)('dbPollMaxInterval', { dataSource })) * 1000,
            exportBucketCsvEscapeSymbol: (0, shared_1.getEnv)('dbExportBucketCsvEscapeSymbol', { dataSource }),
        };
        (0, shared_1.getEnv)('dbExportBucketType', {
            dataSource,
            supported: ['gcp'],
        });
        this.bigquery = new bigquery_1.BigQuery(this.options);
        if (this.options.exportBucket) {
            this.storage = new storage_1.Storage(this.options);
            this.bucket = this.storage.bucket(this.options.exportBucket);
        }
    }
    static driverEnvVariables() {
        // TODO (buntarb): check how this method can/must be used with split
        // names by the data source.
        return [
            'CUBEJS_DB_BQ_PROJECT_ID',
            'CUBEJS_DB_BQ_KEY_FILE',
        ];
    }
    async testConnection() {
        await this.bigquery.query({
            query: 'SELECT ? AS number',
            params: ['1'],
            jobTimeoutMs: this.testConnectionTimeout(),
        });
    }
    readOnly() {
        return !!this.options.readOnly;
    }
    async query(query, values, options) {
        const data = await this.runQueryJob({
            query,
            params: values,
            parameterMode: 'positional',
            useLegacySql: false
        }, options);
        return (data[0] && data[0].map(row => (0, HydrationStream_1.transformRow)(row)));
    }
    async loadTablesForDataset(dataset) {
        try {
            const result = await dataset.query({
                query: `
        SELECT
          columns.column_name as ${this.quoteIdentifier('column_name')},
          columns.table_name as ${this.quoteIdentifier('table_name')},
          columns.table_schema as ${this.quoteIdentifier('table_schema')},
          columns.data_type as ${this.quoteIdentifier('data_type')}
        FROM INFORMATION_SCHEMA.COLUMNS
      `
            });
            if (result.length) {
                return ramda_1.default.reduce(this.informationColumnsSchemaReducer, {}, result[0]);
            }
            return [];
        }
        catch (e) {
            if (e.message.includes('Permission bigquery.tables.get denied on table')) {
                return {};
            }
            throw e;
        }
    }
    async tablesSchema() {
        const dataSets = await this.bigquery.getDatasets();
        const dataSetsColumns = await Promise.all(dataSets[0].map((dataSet) => this.loadTablesForDataset(dataSet)));
        return dataSetsColumns.reduce((prev, current) => Object.assign(prev, current), {});
    }
    async getSchemas() {
        const dataSets = await this.bigquery.getDatasets();
        return dataSets[0].filter((dataSet) => dataSet.id).map((dataSet) => ({
            schema_name: dataSet.id,
        }));
    }
    async getTablesForSpecificSchemas(schemas) {
        try {
            const allTablePromises = schemas.map(async (schema) => {
                const tables = await this.getTablesQuery(schema.schema_name);
                return tables
                    .filter(table => table.table_name)
                    .map(table => ({ schema_name: schema.schema_name, table_name: table.table_name }));
            });
            const allTables = await Promise.all(allTablePromises);
            return allTables.flat();
        }
        catch (e) {
            console.error('Error fetching tables for schemas:', e);
            throw e;
        }
    }
    async getColumnsForSpecificTables(tables) {
        try {
            const allColumnPromises = tables.map(async (table) => {
                const tableName = `${table.schema_name}.${table.table_name}`;
                const columns = await this.tableColumnTypes(tableName);
                return columns.map((column) => ({
                    schema_name: table.schema_name,
                    table_name: table.table_name,
                    data_type: column.type,
                    column_name: column.name,
                }));
            });
            const allColumns = await Promise.all(allColumnPromises);
            return allColumns.flat();
        }
        catch (e) {
            console.error('Error fetching columns for tables:', e);
            throw e;
        }
    }
    async getTablesQuery(schemaName) {
        try {
            const dataSet = await this.bigquery.dataset(schemaName);
            if (!dataSet) {
                return [];
            }
            const [tables] = await this.bigquery.dataset(schemaName).getTables();
            return tables.map(t => ({ table_name: t.id }));
        }
        catch (e) {
            if (e.toString().indexOf('Not found')) {
                return [];
            }
            throw e;
        }
    }
    async tableColumnTypes(table) {
        const [schema, name] = table.split('.');
        const [bigQueryTable] = await this.bigquery.dataset(schema).table(name).getMetadata();
        return bigQueryTable.schema.fields.map((c) => ({ name: c.name, type: this.toGenericType(c.type) }));
    }
    async createSchemaIfNotExists(schemaName) {
        await this.bigquery.dataset(schemaName).get({ autoCreate: true });
    }
    async isUnloadSupported() {
        return this.bucket !== null;
    }
    async stream(query, values) {
        const stream = await this.bigquery.createQueryStream({
            query,
            params: values,
            parameterMode: 'positional',
            useLegacySql: false
        });
        const rowStream = new HydrationStream_1.HydrationStream();
        stream.pipe(rowStream);
        return {
            rowStream,
        };
    }
    async unload(table) {
        if (!this.bucket) {
            throw new Error('Unload is not configured');
        }
        const destination = this.bucket.file(`${table}-*.csv.gz`);
        const [schema, tableName] = table.split('.');
        const bigQueryTable = this.bigquery.dataset(schema).table(tableName);
        const [job] = await bigQueryTable.createExtractJob(destination, { format: 'CSV', gzip: true });
        await this.waitForJobResult(job, { table }, false);
        const [files] = await this.bucket.getFiles({ prefix: `${table}-` });
        const urls = await Promise.all(files.map(async (file) => {
            const [url] = await file.getSignedUrl({
                action: 'read',
                expires: new Date(new Date().getTime() + 60 * 60 * 1000),
            });
            return url;
        }));
        return {
            exportBucketCsvEscapeSymbol: this.options.exportBucketCsvEscapeSymbol,
            csvFile: urls,
        };
    }
    async loadPreAggregationIntoTable(preAggregationTableName, loadSql, params, options) {
        const [dataSet, tableName] = preAggregationTableName.split('.');
        const bigQueryQuery = {
            query: loadSql,
            params,
            parameterMode: 'positional',
            destination: this.bigquery.dataset(dataSet).table(tableName),
            createDisposition: 'CREATE_IF_NEEDED',
            useLegacySql: false
        };
        return this.runQueryJob(bigQueryQuery, options, false);
    }
    async awaitForJobStatus(job, options, withResults) {
        const [result] = await job.getMetadata();
        if (result.status && result.status.state === 'DONE') {
            if (result.status.errorResult) {
                throw new Error(result.status.errorResult.message ?
                    result.status.errorResult.message :
                    JSON.stringify(result.status.errorResult));
            }
            this.reportQueryUsage(result.statistics, options);
        }
        else {
            return null;
        }
        return withResults ? job.getQueryResults() : true;
    }
    async runQueryJob(bigQueryQuery, options, withResults = true) {
        const [job] = await this.bigquery.createQueryJob(bigQueryQuery);
        return this.waitForJobResult(job, options, withResults);
    }
    async waitForJobResult(job, options, withResults) {
        const startedTime = Date.now();
        for (let i = 0; Date.now() - startedTime <= this.options.pollTimeout; i++) {
            const result = await this.awaitForJobStatus(job, options, withResults);
            if (result) {
                return result;
            }
            await (0, shared_1.pausePromise)(Math.min(this.options.pollMaxInterval, 200 * i));
        }
        await job.cancel();
        throw new Error(`BigQuery job timeout reached ${this.options.pollTimeout}ms`);
    }
    quoteIdentifier(identifier) {
        const nestedFields = identifier.split('.');
        return nestedFields.map(name => {
            if (name.match(/^[a-z0-9_]+$/)) {
                return name;
            }
            return `\`${identifier}\``;
        }).join('.');
    }
    capabilities() {
        return {
            incrementalSchemaLoading: true,
        };
    }
}
exports.BigQueryDriver = BigQueryDriver;
//# sourceMappingURL=BigQueryDriver.js.map